
// RCRL Circuit Transient Analysis
// ===============================
// Show step response of the circuit
// Example for course: IMS
// Method: Runge-Kutta 4
// Warning: Needs linking with math library (GCC option -lm)
// Copyright: (c) Petr Peringer, FIT TU Brno, 2020


#include <stdio.h>
#include <stdlib.h>     // strtod()

// System:
#define SIZE 2                  // state vector size of the system
const double R1 = 90;           // Ohm
const double R2 = 10;           // Ohm
const double C = 100e-6;        // F
const double L = 100e-3;        // H
const double uin = 1.0;         // step 0->1V at time 0

// "Dynamic section" = continuous system description
// Input:  t = time of step start,
//         y = state vector = integrator outputs
// Output: yin = vector of derivatives = integrators input values
void f(double t, double *y, double *yin)
{
    // State variables (assign readable names)
    double u_C = y[0];
    double i_L = y[1];
    // Equations transformed to assignments:
    double u_R2 = R2 * i_L;
    double u_R1 = uin - u_C;    // uin==u_1
    double u_L = u_C - u_R2;
    double i = u_R1 / R1;
    double i_C = i - i_L;
    yin[0] = i_C/C;             // der(u_C)
    yin[1] = u_L/L;             // der(i_L)
}

// Runge-Kutta-4 step
void RK4step(double t, double h, double *y);

// Experiment control
int main(int argc, char *argv[])
{
    const double TEND = 0.08;
    double y[SIZE] = { 0, 0 };               // initial state
    const double stepsize = 0.00001;         // cca 2000/period
    const unsigned NS = (TEND/stepsize)/500; // print interval
    printf("# R1=%g R2=%g L=%g C=%g\n", R1, R2, L, C);

    unsigned steps = 0;
    double t = 0;
    while (t < TEND) {
        double u_C = y[0];
        double i_L = y[1];
        double u_R2 = R2 * i_L;
        double u_L = u_C - u_R2;
        if(steps%NS == 0)               // output each NS steps
            printf("%8.6f % -11g % -11g  % -11g\n", t, u_C, u_L, i_L);
        RK4step(t, stepsize, y);        // make step
        t += stepsize;          // advance the simulation time
        ++steps;
    }
    return 0;
}

//////////////////////////////////////////////////////////////////
// Runge-Kutta 4-th order method step
// Parameters: t = start time,
//             h = step size,
//             y = state vector
// Output:  new state in y
// Warning: it does not update time
void RK4step(double t, double h, double *y)
{
    double yin[SIZE];           // integrator input = derivative
    double ystart[SIZE];        // initial state
    // 4 stages results:
    double k1[SIZE];
    double k2[SIZE];
    double k3[SIZE];
    double k4[SIZE];

    int i;
    for (i = 0; i < SIZE; i++)  // save initial value
        ystart[i] = y[i];

    f(t, y, yin);               // yin = f(t, y(t))

    for (i = 0; i < SIZE; i++) {
        k1[i] = h * yin[i];     // k1 = h * f(t, y(t))
        y[i] = ystart[i] + k1[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k1/2)

    for (i = 0; i < SIZE; i++) {
        k2[i] = h * yin[i];     // k2 = h * f(t+h/2, y(t)+k1/2)
        y[i] = ystart[i] + k2[i] / 2;
    }

    f(t + h / 2, y, yin);       // yin = f(t+h/2, y(t)+k2/2)

    for (i = 0; i < SIZE; i++) {
        k3[i] = h * yin[i];     // k3 = h * f(t+h/2, y(t)+k2/2)
        y[i] = ystart[i] + k3[i];
    }

    f(t + h, y, yin);           // yin = f(t+h, y(t)+k3)

    for (i = 0; i < SIZE; i++) {
        k4[i] = h * yin[i];     // k4 = h * f(t+h, y(t)+k3)
        // Result:
        // y(t+h) = y(t) + k1/6 + k2/3 + k3/3 + k4/6;
        y[i] = ystart[i] + k1[i] / 6 + k2[i] / 3 + k3[i] / 3 + k4[i] / 6;
    }
    // Return: y = new state at time t+h
}
